import unittest

from pystan import stanc, StanModel
from pystan._compat import PY2

class TestUTF8(unittest.TestCase):

    desired = sorted({"status", "model_cppname", "cppcode", "model_name", "model_code", "include_paths"})

    def test_utf8(self):
        model_code = 'parameters {real y;} model {y ~ normal(0,1);}'
        result = stanc(model_code=model_code)
        self.assertEqual(sorted(result.keys()), self.desired)
        self.assertTrue(result['cppcode'].startswith("// Code generated by Stan "))
        self.assertEqual(result['status'], 0)

    def test_utf8_linecomment(self):
        model_code = u'parameters {real y;\n //äöéü\n} model {y ~ normal(0,1);}'
        result = stanc(model_code=model_code)
        self.assertEqual(sorted(result.keys()), self.desired)
        self.assertTrue(result['cppcode'].startswith("// Code generated by Stan "))
        self.assertEqual(result['status'], 0)

    def test_utf8_multilinecomment(self):
        model_code = u'parameters {real y;\n /*äöéü\näöéü*/\n} model {y ~ normal(0,1);}'
        result = stanc(model_code=model_code)
        self.assertEqual(sorted(result.keys()), self.desired)
        self.assertTrue(result['cppcode'].startswith("// Code generated by Stan "))
        self.assertEqual(result['status'], 0)

    def test_utf8_inprogramcode(self):
        model_code = u'parameters {real ö;\n} model {ö ~ normal(0,1);}'
        assertRaisesRegex = self.assertRaisesRegexp if PY2 else self.assertRaisesRegex
        with assertRaisesRegex(ValueError, 'Failed to parse Stan model .*'):
            stanc(model_code=model_code)



# pymc git commit: 6115726122d46267c86d16de635941daa37eb357
# =======
# License
# =======
#
# PyMC is distributed under the Apache License, Version 2.0
#
# Copyright (c) 2006 Christopher J. Fonnesbeck (Academic Free License)
# Copyright (c) 2007-2008 Christopher J. Fonnesbeck, Anand Prabhakar Patil, David Huard (Academic Free License)
# Copyright (c) 2009-2013 The PyMC developers (see contributors to pymc-devs on GitHub)
# All rights reserved.

from pylab import *
try:
    import matplotlib.gridspec as gridspec
except ImportError:
    gridspec = None
import numpy as np
from scipy.stats import kde
from .stats import *
from .trace import *

__all__ = ['traceplot', 'kdeplot', 'kde2plot', 'forestplot', 'autocorrplot']

def traceplot(trace, vars=None):
    if vars is None:
        vars = trace.varnames

    if isinstance(trace, MultiTrace):
        trace = trace.combined()

    n = len(vars)
    f, ax = subplots(n, 2, squeeze=False)

    for i, v in enumerate(vars):
        d = np.squeeze(trace[v])

        if trace[v].dtype.kind == 'i':
            ax[i, 0].hist(d, bins=sqrt(d.size))
        else:
            kdeplot_op(ax[i, 0], d)
        ax[i, 0].set_title(str(v))
        ax[i, 1].plot(d, alpha=.35)

        ax[i, 0].set_ylabel("frequency")
        ax[i, 1].set_ylabel("sample value")

    return f


def kdeplot_op(ax, data):
    data = np.atleast_2d(data.T).T
    for i in range(data.shape[1]):
        d = data[:, i]
        density = kde.gaussian_kde(d)
        l = np.min(d)
        u = np.max(d)
        x = np.linspace(0, 1, 100) * (u - l) + l

        ax.plot(x, density(x))


def kde2plot_op(ax, x, y, grid=200):
    xmin = x.min()
    xmax = x.max()
    ymin = y.min()
    ymax = y.max()

    grid = grid * 1j
    X, Y = np.mgrid[xmin:xmax:grid, ymin:ymax:grid]
    positions = np.vstack([X.ravel(), Y.ravel()])
    values = np.vstack([x, y])
    kernel = kde.gaussian_kde(values)
    Z = np.reshape(kernel(positions).T, X.shape)

    ax.imshow(np.rot90(Z), cmap=cm.gist_earth_r,
              extent=[xmin, xmax, ymin, ymax])


def kdeplot(data):
    f, ax = subplots(1, 1, squeeze=True)
    kdeplot_op(ax, data)
    return f


def kde2plot(x, y, grid=200):
    f, ax = subplots(1, 1, squeeze=True)
    kde2plot_op(ax, x, y, grid)
    return f

def autocorrplot(trace, vars=None, fontmap = None, max_lag=100):
    """Bar plot of the autocorrelation function for a trace"""

    try:
        # MultiTrace
        traces = trace.traces

    except AttributeError:

        # NpTrace
        traces = [trace]

    if fontmap is None: fontmap = {1:10, 2:8, 3:6, 4:5, 5:4}

    if vars is None:
        vars = traces[0].varnames

    # Extract sample data
    samples = [{v:trace[v] for v in vars} for trace in traces]

    chains = len(traces)

    n = len(samples[0])
    f, ax = subplots(n, chains, squeeze=False)

    max_lag = min(len(samples[0][vars[0]])-1, max_lag)

    for i, v in enumerate(vars):

        for j in xrange(chains):

            d = np.squeeze(samples[j][v])

            ax[i,j].acorr(d, detrend=mlab.detrend_mean, maxlags=max_lag)

            if not j:
                ax[i, j].set_ylabel("correlation")
            ax[i, j].set_xlabel("lag")

            if chains > 1:
                ax[i, j].set_title("chain {0}".format(j+1))

    # Smaller tick labels
    tlabels = gca().get_xticklabels()
    setp(tlabels, 'fontsize', fontmap[1])

    tlabels = gca().get_yticklabels()
    setp(tlabels, 'fontsize', fontmap[1])


def var_str(name, shape):
    """Return a sequence of strings naming the element of the tallyable object.
    This is a support function for forestplot.

    :Example:
    >>> var_str('theta', (4,))
    ['theta[1]', 'theta[2]', 'theta[3]', 'theta[4]']

    """

    size = prod(shape)
    ind = (indices(shape) + 1).reshape(-1, size)
    names = ['[' + ','.join(map(str, i)) + ']' for i in zip(*ind)]
    # if len(name)>12:
    #     name = '\n'.join(name.split('_'))
    #     name += '\n'
    names[0] = '%s %s' % (name, names[0])
    return names


def forestplot(trace_obj, vars=None, alpha=0.05, quartiles=True, rhat=True,
               main=None, xtitle=None, xrange=None, ylabels=None, chain_spacing=0.05, vline=0):
    """ Forest plot (model summary plot)

    Generates a "forest plot" of 100*(1-alpha)% credible intervals for either the
    set of variables in a given model, or a specified set of nodes.

    :Arguments:
        trace_obj: NpTrace or MultiTrace object
            Trace(s) from an MCMC sample.

        vars: list
            List of variables to plot (defaults to None, which results in all
            variables plotted).

        alpha (optional): float
            Alpha value for (1-alpha)*100% credible intervals (defaults to 0.05).

        quartiles (optional): bool
            Flag for plotting the interquartile range, in addition to the
            (1-alpha)*100% intervals (defaults to True).

        rhat (optional): bool
            Flag for plotting Gelman-Rubin statistics. Requires 2 or more
            chains (defaults to True).

        main (optional): string
            Title for main plot. Passing False results in titles being
            suppressed; passing None (default) results in default titles.

        xtitle (optional): string
            Label for x-axis. Defaults to no label

        xrange (optional): list or tuple
            Range for x-axis. Defaults to matplotlib's best guess.

        ylabels (optional): list
            User-defined labels for each variable. If not provided, the node
            __name__ attributes are used.

        chain_spacing (optional): float
            Plot spacing between chains (defaults to 0.05).

        vline (optional): numeric
            Location of vertical reference line (defaults to 0).

    """

    if not gridspec:
        print_(
            '\nYour installation of matplotlib is not recent enough to support summary_plot; this function is disabled until matplotlib is updated.')
        return

    # Quantiles to be calculated
    qlist = [100 * alpha / 2, 50, 100 * (1 - alpha / 2)]
    if quartiles:
        qlist = [100 * alpha / 2, 25, 50, 75, 100 * (1 - alpha / 2)]

    # Range for x-axis
    plotrange = None

    # Number of chains
    chains = None

    # Gridspec
    gs = None

    # Subplots
    interval_plot = None
    rhat_plot = None

    try:
        # First try MultiTrace type
        traces = trace_obj.traces

        if rhat and len(traces) > 1:

            from .diagnostics import gelman_rubin

            R = gelman_rubin(trace_obj)
            if vars is not None:
                R = {v: R[v] for v in vars}

        else:

            rhat = False

    except AttributeError:

        # Single NpTrace
        traces = [trace_obj]

        # Can't calculate Gelman-Rubin with a single trace
        rhat = False

    if vars is None:
        vars = traces[0].varnames

    # Empty list for y-axis labels
    labels = []

    chains = len(traces)

    if gs is None:
        # Initialize plot
        if rhat and chains > 1:
            gs = gridspec.GridSpec(1, 2, width_ratios=[3, 1])

        else:

            gs = gridspec.GridSpec(1, 1)

        # Subplot for confidence intervals
        interval_plot = subplot(gs[0])

    for j, tr in enumerate(traces):

        # Get quantiles
        trace_quantiles = quantiles(tr, qlist)
        hpd_intervals = hpd(tr, alpha)

        # Counter for current variable
        var = 1

        for varname in vars:

            var_quantiles = trace_quantiles[varname]

            quants = var_quantiles.values()
            var_hpd = hpd_intervals[varname].T

            # Substitute HPD interval for quantile
            quants[0] = var_hpd[0].T
            quants[-1] = var_hpd[1].T

            # Ensure x-axis contains range of current interval
            if plotrange:
                plotrange = [min(
                             plotrange[0],
                             np.min(quants)),
                             max(plotrange[1],
                                 np.max(quants))]
            else:
                plotrange = [np.min(quants), np.max(quants)]

            # Number of elements in current variable
            value = tr[varname][0]
            k = np.size(value)

            # Append variable name(s) to list
            if not j:
                if k > 1:
                    names = var_str(varname, shape(value))
                    labels += names
                else:
                    labels.append(varname)
                    # labels.append('\n'.join(varname.split('_')))

            # Add spacing for each chain, if more than one
            e = [0] + [(chain_spacing * ((i + 2) / 2)) *
                (-1) ** i for i in range(chains - 1)]

            # Deal with multivariate nodes
            if k > 1:

                for i, q in enumerate(np.transpose(quants).squeeze()):

                    # Y coordinate with jitter
                    y = -(var + i) + e[j]

                    if quartiles:
                        # Plot median
                        plot(q[2], y, 'bo', markersize=4)
                        # Plot quartile interval
                        errorbar(
                            x=(q[1],
                                q[3]),
                            y=(y,
                                y),
                            linewidth=2,
                            color="blue")

                    else:
                        # Plot median
                        plot(q[1], y, 'bo', markersize=4)

                    # Plot outer interval
                    errorbar(
                        x=(q[0],
                            q[-1]),
                        y=(y,
                            y),
                        linewidth=1,
                        color="blue")

            else:

                # Y coordinate with jitter
                y = -var + e[j]

                if quartiles:
                    # Plot median
                    plot(quants[2], y, 'bo', markersize=4)
                    # Plot quartile interval
                    errorbar(
                        x=(quants[1],
                            quants[3]),
                        y=(y,
                            y),
                        linewidth=2,
                        color="blue")
                else:
                    # Plot median
                    plot(quants[1], y, 'bo', markersize=4)

                # Plot outer interval
                errorbar(
                    x=(quants[0],
                        quants[-1]),
                    y=(y,
                        y),
                    linewidth=1,
                    color="blue")

            # Increment index
            var += k

    labels = ylabels or labels

    # Update margins
    left_margin = np.max([len(x) for x in labels]) * 0.015
    gs.update(left=left_margin, right=0.95, top=0.9, bottom=0.05)

    # Define range of y-axis
    ylim(-var + 0.5, -0.5)

    datarange = plotrange[1] - plotrange[0]
    xlim(plotrange[0] - 0.05 * datarange, plotrange[1] + 0.05 * datarange)

    # Add variable labels
    yticks([-(l + 1) for l in range(len(labels))], labels)

    # Add title
    if main is not False:
        plot_title = main or str(int((
            1 - alpha) * 100)) + "% Credible Intervals"
        title(plot_title)

    # Add x-axis label
    if xtitle is not None:
        xlabel(xtitle)

    # Constrain to specified range
    if xrange is not None:
        xlim(*xrange)

    # Remove ticklines on y-axes
    for ticks in interval_plot.yaxis.get_major_ticks():
        ticks.tick1On = False
        ticks.tick2On = False

    for loc, spine in interval_plot.spines.iteritems():
        if loc in ['bottom', 'top']:
            pass
            # spine.set_position(('outward',10)) # outward by 10 points
        elif loc in ['left', 'right']:
            spine.set_color('none')  # don't draw spine

    # Reference line
    axvline(vline, color='k', linestyle='--')

    # Genenerate Gelman-Rubin plot
    if rhat and chains > 1:

        # If there are multiple chains, calculate R-hat
        rhat_plot = subplot(gs[1])

        if main is not False:
            title("R-hat")

        # Set x range
        xlim(0.9, 2.1)

        # X axis labels
        xticks((1.0, 1.5, 2.0), ("1", "1.5", "2+"))
        yticks([-(l + 1) for l in range(len(labels))], "")

        i = 1
        for varname in vars:

            value = traces[0][varname][0]
            k = np.size(value)

            if k > 1:
                plot([min(r, 2) for r in R[varname]], [-(j + i)
                     for j in range(k)], 'bo', markersize=4)
            else:
                plot(min(R[varname], 2), -i, 'bo', markersize=4)

            i += k

        # Define range of y-axis
        ylim(-i + 0.5, -0.5)

        # Remove ticklines on y-axes
        for ticks in rhat_plot.yaxis.get_major_ticks():
            ticks.tick1On = False
            ticks.tick2On = False

        for loc, spine in rhat_plot.spines.iteritems():
            if loc in ['bottom', 'top']:
                pass
                # spine.set_position(('outward',10)) # outward by 10 points
            elif loc in ['left', 'right']:
                spine.set_color('none')  # don't draw spine

    return gs



#-----------------------------------------------------------------------------
# Copyright (c) 2013-2015 PyStan developers
#
# This file is licensed under Version 3.0 of the GNU General Public
# License. See LICENSE for a text of the license.
#-----------------------------------------------------------------------------
import logging

from pystan.api import stanc, stan
from pystan.misc import read_rdump, stan_rdump, stansummary
from pystan.diagnostics import check_hmc_diagnostics
from pystan.model import StanModel
from pystan.lookup import lookup

logger = logging.getLogger('pystan')
logger.addHandler(logging.NullHandler())
if len(logger.handlers) == 1:
    logging.basicConfig(level=logging.INFO)

# following PEP 386
# See also https://docs.openstack.org/pbr/latest/user/semver.html
__version__ = '2.19.0.0'



import unittest

import numpy as np

import pystan
from pystan._compat import PY2


class TestFixedParam(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        model_code = '''
            model { }
            generated quantities {
              real y;
              y = normal_rng(0, 1);
            }
        '''
        cls.model = pystan.StanModel(model_code=model_code)

    def test_fixed_param(self):
        fit = self.model.sampling(algorithm='Fixed_param')
        self.assertEqual(fit.sim['iter'], 2000)
        self.assertEqual(fit.sim['pars_oi'], ['y', 'lp__'])
        self.assertEqual(len(fit.sim['samples']), 4)
        for sample_dict in fit.sim['samples']:
            assert -0.5 < np.mean(sample_dict['chains']['y']) < 0.5
            assert 0.5 < np.std(sample_dict['chains']['y']) < 1.5



#-----------------------------------------------------------------------------
# Copyright (c) 2017, PyStan developers
#
# This file is licensed under Version 3.0 of the GNU General Public
# License. See LICENSE for a text of the license.
#-----------------------------------------------------------------------------

import numpy as np
import re
import pkg_resources
import io

lookuptable = None
stanftable = None

def lookup(name, min_similarity_ratio=.75):
    """
    Look up for a Stan function with similar functionality to a Python
    function (or even an R function, see examples). If the function is
    not present on the lookup table, then attempts to find similar one
    and prints the results. This function requires package `pandas`.

    Parameters
    -----------
    name : str
        Name of the function one wants to look for.
    min_similarity_ratio : float
        In case no exact match is found on the lookup table, the
        function will attempt to find similar names using
        `difflib.SequenceMatcher.ratio()`, and then results with
        calculated ratio below `min_similarity_ratio` will be discarded.

    Examples
    ---------
    #Look up for a Stan function similar to scipy.stats.skewnorm
    lookup("scipy.stats.skewnorm")
    #Look up for a Stan function similar to R dnorm
    lookup("R.dnorm")
    #Look up for a Stan function similar to numpy.hstack
    lookup("numpy.hstack")
    #List Stan log probability mass functions
    lookup("lpmfs")
    #List Stan log cumulative density functions
    lookup("lcdfs")

    Returns
    ---------
    A pandas.core.frame.DataFrame if exact or at least one similar
    result is found, None otherwise.
    """
    if lookuptable is None:
        build()
    if name not in lookuptable.keys():
        from difflib import SequenceMatcher
        from operator import itemgetter
        print("No match for " + name + " in the lookup table.")

        lkt_keys = list(lookuptable.keys())
        mapfunction = lambda x: SequenceMatcher(a=name, b=x).ratio()
        similars = list(map(mapfunction, lkt_keys))
        similars = zip(range(len(similars)), similars)
        similars = list(filter(lambda x: x[1] >= min_similarity_ratio,
                               similars))
        similars = sorted(similars, key=itemgetter(1))

        if (len(similars)):
            print("But the following similar entries were found: ")
            for i in range(len(similars)):
                print(lkt_keys[similars[i][0]] + " ===> with similary "
                      "ratio of " + str(round(similars[i][1], 3)) + "")
            print("Will return results for entry"
                  " " + lkt_keys[similars[i][0]] + " "
                  "(which is the most similar entry found).")
            return lookup(lkt_keys[similars[i][0]])
        else:
            print("And no similar entry found. You may try to decrease"
                  "the min_similarity_ratio parameter.")
        return
    entries = stanftable[lookuptable[name]]
    if not len(entries):
        return "Found no equivalent Stan function available for " + name

    try:
        import pandas as pd
    except ImportError:
        raise ImportError('Package pandas is require to use this '
                          'function.')

    return pd.DataFrame(entries)



def build():
    def load_table_file(fname):
        fname = "lookuptable/" + fname
        fbytes = pkg_resources.resource_string(__name__, fname)
        return io.BytesIO(fbytes)
    stanfunctions_file = load_table_file("stan-functions.txt")
    rfunctions_file = load_table_file("R.txt")
    pythontb_file = load_table_file("python.txt")

    stanftb = np.genfromtxt(stanfunctions_file, delimiter=';',
                            names=True, skip_header=True,
                            dtype=['<U200','<U200','<U200' ,"int"])
    rpl_textbar = np.vectorize(lambda x: x.replace("\\textbar \\", "|"))
    stanftb['Arguments'] = rpl_textbar(stanftb['Arguments'])

    StanFunction = stanftb["StanFunction"]

    #Auto-extract R functions
    rmatches = [re.findall(r'('
                           '(?<=RFunction\[StanFunction == \").+?(?=\")'
                           '|(?<=grepl\(").+?(?=", StanFunction\))'
                           '|(?<= \<\- ").+?(?="\)))'
                           '|NA\_character\_', l.decode("utf-8"))
                for l in rfunctions_file]
    tomatch = list(filter(lambda x: len(x) == 2, rmatches))
    tomatch = np.array(tomatch, dtype=str)
    tomatch[:, 1] = np.vectorize(lambda x: "R." + x)(tomatch[:,1])

    #Get packages lookup table for Python packages
    pymatches = np.genfromtxt(pythontb_file, delimiter='; ', dtype=str)
    tomatch = np.vstack((tomatch, pymatches))

    lookuptb = dict()
    for i in range(tomatch.shape[0]):
        matchedlines = np.vectorize(lambda x: re.match(tomatch[i, 0],
                                    x))(StanFunction)
        lookuptb[tomatch[i, 1]] = np.where(matchedlines)[0]

    #debug: list of rmatches that got wrong
    #print(list(filter(lambda x: len(x) != 2 and len(x) != 0,
    #                  rmatches)))

    #debug: list of nodes without matches on lookup table
    #for k in lookuptb:
    #    if len(lookuptb[k]) == 0:
    #        print(k)
    global lookuptable
    global stanftable

    stanftable = stanftb
    lookuptable = lookuptb



# distutils: language = c++

from libcpp cimport bool
from libcpp.map cimport map
from libcpp.pair cimport pair
from libcpp.string cimport string
from libcpp.vector cimport vector


cdef extern from "stan/io/var_context.hpp" namespace "stan::io":
    cdef cppclass var_context:
        bool contains_r(const string&)
        bool contains_i(const string&)
        vector[double] vals_r(const string&)
        vector[size_t] dims_r(const string&)
        vector[int] vals_i(const string&)
        vector[size_t] dims_i(const string&)
        void names_r(vector[string]&)
        void names_i(vector[string]&)
        bool remove(const string&)


cdef extern from "py_var_context.hpp" namespace "pystan::io":
    cdef cppclass py_var_context(var_context):
        py_var_context(map[string, pair[vector[double], vector[size_t]]]&,
                       map[string, pair[vector[int], vector[size_t]]]&)




import pystan._chains as _chains
from numpy import nan

def ess(sim, n):
    """Calculate effective sample size

    Parameters
    ----------
    sim : chains
    n : int
        Parameter index starting from 0
    """
    try:
        ess = _chains.effective_sample_size(sim, n)
    except (ValueError, ZeroDivisionError):
        ess = nan
    return ess


def splitrhat(sim, n):
    """Calculate rhat

    Parameters
    ----------
    sim : chains
    n : int
        Parameter index starting from 0
    """
    try:
        rhat = _chains.split_potential_scale_reduction(sim, n)
    except (ValueError, ZeroDivisionError):
        rhat = nan
    return rhat

def ess_and_splitrhat(sim, n):
    """Calculate ess and rhat

    This saves time by creating just one stan::mcmc::chains instance.
    """
    # FIXME: does not yet save time
    return (ess(sim, n), splitrhat(sim, n))



#!/usr/bin/python

"""
Replacement for runtest target in Makefile.

Call script with '-h' as an option to see a helpful message.
"""

from __future__ import print_function
from argparse import ArgumentParser, RawTextHelpFormatter
import os
import os.path
import platform
import subprocess
import sys
import time

winsfx = ".exe"
testsfx = "_test.cpp"


def processCLIArgs():
    """
    Define and process the command line interface to the runTests.py script.
    """
    cli_description = "Generate and run stan math library tests."
    cli_epilog = "See more information at: https://github.com/stan-dev/math"

    parser = ArgumentParser(description=cli_description,
                            epilog=cli_epilog,
                            formatter_class=RawTextHelpFormatter)

    # Now define all the rules of the command line args and opts
    parser.add_argument("-j", metavar="N", type=int, default=1,
                        help="number of cores for make to use")

    tests_help_msg = "The path(s) to the test case(s) to run.\n"
    tests_help_msg += "Example: 'test/unit', 'test/prob', and/or\n"
    tests_help_msg += "         'test/unit/math/prim/scal/fun/abs_test.cpp'"
    parser.add_argument("tests", nargs="+", type=str,
                        help=tests_help_msg)
    f_help_msg = "Only tests with file names matching these will be executed.\n"
    f_help_msg += "Example: '-f chol', '-f opencl', '-f prim'"
    parser.add_argument("-f", type=str, default = [], action="append",
                        help=f_help_msg)
    parser.add_argument("-d", "--debug", dest="debug", action="store_true",
                        help="request additional script debugging output.")
    parser.add_argument("-m", "--make-only", dest="make_only",
                        action="store_true", help="Don't run tests, just try to make them.")
    parser.add_argument("--run-all", dest="run_all", action="store_true",
                        help="Don't stop at the first test failure, run all of them.")

    # And parse the command line against those rules
    return parser.parse_args()


def stopErr(msg, returncode):
    """Report an error message to stderr and exit with a given code."""
    sys.stderr.write('%s\n' % msg)
    sys.stderr.write('exit now (%s)\n' % time.strftime('%x %X %Z'))
    sys.exit(returncode)


def isWin():
    return (platform.system().lower().startswith("windows")
            or os.name.lower().startswith("windows"))


batchSize = 20 if isWin() else 200


def mungeName(name):
    """Set up the makefile target name"""
    if (name.endswith(testsfx)):
        name = name.replace(testsfx, "_test")

        if (isWin()):
            name += winsfx
            name = name.replace("\\", "/")

    return name

def doCommand(command, exit_on_failure=True):
    """Run command as a shell command and report/exit on errors."""
    print("------------------------------------------------------------")
    print("%s" % command)
    p1 = subprocess.Popen(command, shell=True)
    p1.wait()
    if exit_on_failure and (not(p1.returncode is None) and not(p1.returncode == 0)):
        stopErr('%s failed' % command, p1.returncode)


def generateTests(j):
    """Generate all tests and pass along the j parameter to make."""
    doCommand('make -j%d generate-tests -s' % (j or 1))


def makeTest(name, j):
    """Run the make command for a given single test."""
    doCommand('make -j%d %s' % (j or 1, name))

def commandExists(command):
    p = subprocess.Popen(command, shell=True,
                         stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)
    p.wait()
    return p.returncode != 127

def runTest(name, run_all=False, mpi=False, j=1):
    executable = mungeName(name).replace("/", os.sep)
    xml = mungeName(name).replace(winsfx, "")
    command = '%s --gtest_output="xml:%s.xml"' % (executable, xml)
    if mpi:
        if not commandExists("mpirun"):
            stopErr("Error: need to have mpi (and mpirun) installed to run mpi tests"
                    + "\nCheck https://github.com/stan-dev/stan/wiki/Parallelism-using-MPI-in-Stan for more details."
                    , -1)
        if "mpi_" in name:
            j = j > 2 and j or 2
        else:
            j = 1
        command = "mpirun -np {} {}".format(j, command)
    doCommand(command, not run_all)

def findTests(base_path, filter_names):
    folders = filter(os.path.isdir, base_path)
    nonfolders = list(set(base_path) - set(folders))
    tests = nonfolders + [os.path.join(root, n)
            for f in folders
            for root, _, names in os.walk(f)
            for n in names
            if n.endswith(testsfx)]
    tests = map(mungeName, tests)
    tests = [test
            for test in tests
            if all(filter_name in test
                   for filter_name in filter_names)]
    return tests

def batched(tests):
    return [tests[i:i + batchSize] for i in range(0, len(tests), batchSize)]


def main():
    inputs = processCLIArgs()

    try:
        with open("make/local") as f:
            stan_mpi =  "STAN_MPI" in f.read()
    except IOError:
        stan_mpi = False

    # pass 0: generate all auto-generated tests
    if any(['test/prob' in arg for arg in inputs.tests]):
        generateTests(inputs.j)

    tests = findTests(inputs.tests, inputs.f)
    if not tests:
        stopErr("No matching tests found.", -1)
    if inputs.debug:
        print("Collected the following tests:\n", tests)

    # pass 1: make test executables
    for batch in batched(tests):
        if inputs.debug:
            print("Test batch: ", batch)
        makeTest(" ".join(batch), inputs.j)

    if not inputs.make_only:
        # pass 2: run test targets
        for t in tests:
            if inputs.debug:
                print("run single test: %s" % testname)
            runTest(t, inputs.run_all, mpi = stan_mpi, j = inputs.j)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
py2/py3 compatibility support drawn from jinja2

see http://lucumr.pocoo.org/2013/5/21/porting-to-python-3-redux/

"""

import sys

PY2 = sys.version_info[0] == 2
_identity = lambda x: x

if not PY2:
    unichr = chr
    range_type = range
    text_type = str
    string_types = (str,)

    implements_to_string = _identity

    ifilter = filter
    imap = map
    izip = zip

else:
    unichr = unichr
    text_type = unicode
    range_type = xrange
    string_types = (str, unicode)

    def implements_to_string(cls):
        cls.__unicode__ = cls.__str__
        cls.__str__ = lambda x: x.__unicode__().encode('utf-8')
        return cls

    from itertools import imap, izip, ifilter
